# -*- coding: utf-8
import re
from os import path

import six
import yaml

NBSP = u'\xa0'


SPECIAL_CHARACTERS_MAP = {  # defines a key-value for unicode replacement characters, used in YAML
    '_': NBSP               # Test_Phrase maps to Test&nbsp;Phrase
}


class Rule(object):
    flags = []

    def __init__(self, pattern=None, replacement=None, flags=[], specs=[]):
        if flags is not None:
            self.flags = flags

        if pattern is not None:
            self.pattern = pattern

        if replacement is not None:
            self.replacement = replacement

        self.replacement = self._translate_special_chars(replacement)

        self._re = None
        self.specs = specs

    def apply(self, text):
        if not self._re:
            self._compile()

        return self._re.sub(self.replacement, text)

    def _compile(self):
        """
        One-time compile of the regex
        """
        resulting_re_flags = 0

        if six.PY2:
            resulting_re_flags |= re.UNICODE

        if len(self.flags):
            for flag in self.flags:
                resulting_re_flags |= getattr(re, flag)

        self._re = re.compile(self.pattern, flags=resulting_re_flags)

    def _translate_special_chars(self, text):
        """
        Unicode non-breaking spaces are marked as '_' in the config
        """
        for p, r in six.iteritems(SPECIAL_CHARACTERS_MAP):
            text = text.replace(p, r)

        return text


class ABRule(Rule):
    """
    A special rule for testing, replaces all a's in text by b's
    """
    pattern = 'a'
    replacement = 'b'


def load_from_file(ruledef):
    """
    Load rules from file
    """
    file = path.join(path.dirname(__file__), 'rules', ruledef + '.yaml')
    with open(file, 'rb') as f:
        for rule_name, rule in six.iteritems(yaml.load(f)):
            yield rule_name, Rule(
                pattern=rule['pattern'],
                replacement=rule['replacement'],
                flags=rule.get('flags'),
                specs=rule.get('specs')
            )



# -*- coding: utf-8

from richtypo import Richtypo
from richtypo.rules import load_from_file

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


def test_rule_loader():
    rules = dict(load_from_file('generic'))

    assert len(rules.keys()) >= 1

    rule = rules['cleanup_before']
    assert rule.pattern == '\s+'
    assert rule.replacement == ' '


def test_rule_loader_with_non_breaking_spaces():
    rules = dict(load_from_file('generic'))

    nbsp = rules['nbsp']
    assert nbsp.replacement == u'\xa0'


@patch('richtypo.Richtypo._get_ruleset_rules')
def test_loading_generic_ruledef(rules):
    """
    If not specified, all rules should be taken from the generic ruledef (rules/generic.yaml)
    """
    rules.return_value = ['one', 'two']

    with patch('richtypo.Richtypo.build_rule_chain'):  # dont actualy build rule chain
        with patch('richtypo.Richtypo._load_rules_from_file') as loader:
            Richtypo()
            assert loader.call_count == 1
            loader.assert_called_with('generic')


@patch('richtypo.Richtypo._get_ruleset_rules')
def test_loading_specified_ruledef(get_ruleset):
    """
    Check for loading specified ruledef

    ruledefs are specified when constructing a ruleset, like this:
        ruledef: rulename

    """
    get_ruleset.return_value = ['ru: test']

    with patch('richtypo.Richtypo.build_rule_chain'):  # dont actualy build rule chain
        with patch('richtypo.Richtypo._load_rules_from_file') as loader:
            Richtypo()
            assert loader.call_count == 1
            loader.assert_called_with('ru')


@patch('richtypo.Richtypo._get_ruleset_rules')
def test_loading_ruledef_only_once(get_ruleset):
    """
    Check if ruledefs are loaded only once
    """
    get_ruleset.return_value = ['ru:test', 'tst-ruledef:test', 'ru:test1', 'ru:test2']

    with patch('richtypo.Richtypo.build_rule_chain'):  # dont actualy build rule chain
        with patch('richtypo.Richtypo._load_rules_from_file') as loader:
            Richtypo()
            assert loader.call_count == 2  #
            loader.assert_any_call('ru')
            loader.assert_any_call('tst-ruledef')



import richtypo

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


__all__ = ['filter']


@lru_cache(maxsize=128)
def _get_typograph(ruleset):
    return richtypo.Richtypo(ruleset)


def filter(value, ruleset='generic'):
    r = _get_typograph(ruleset)
    return r.richtypo(value)




def pytest_generate_tests(metafunc):
    """
    Generate tests for ruledefs in yaml files according to their specs defined in-place
    """
    if 'rule_name' in metafunc.fixturenames:
        from richtypo.rules import load_from_file
        rules = []
        for ruledef in ['generic', 'ru', 'en']:
            rules += [(name, rule) for name, rule in load_from_file(ruledef)]

        metafunc.parametrize('rule_name, rule', rules)



# -*- coding: utf-8
from richtypo.filter import filter

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@patch('richtypo.filter.richtypo.Richtypo._get_ruleset_rules')
def test_filter_caching(rules):
    rules.return_value = []
    with patch('richtypo.filter.richtypo.Richtypo.load_rules_for_ruleset') as loader:
        for i in range(0, 10):
            filter(u'a -- b', 'some-ruleset') == u'a — b'

        assert loader.call_count == 1  # rules should be loaded from file only once


def test_filter():
    from richtypo.rules import NBSP
    got = filter(u'Из-под топота копыт', 'ru-lite')
    assert got == u'Из-под%sтопота копыт' % NBSP


def test_filter_with_default_ruleset():
    got = filter(u'a -- b')
    assert got == u'a — b'



# -*- coding: utf-8

import re

import pytest

from richtypo import Richtypo


@pytest.mark.parametrize("input, expected", [
    ('<h1>test</h1>', '~1~test~2~'),
    ('<h1>тест</h1>', '~1~тест~2~'),
    (r'<h1>test</', r'~1~test</'),
    ('тест', 'тест'),
    ('<!-- h1 -->test</-->', '~1~test~2~'),
    ('<H1>test</H1>', '~1~test~2~'),
    ('<h1>test<h1>', '~1~test~2~'),
    ('<h1><1>test</h1>', '~1~~2~test~3~'),
    ('<p>test is a <a class="testclass">test link</a></p>', '~1~test is a ~2~test link~3~~4~'),
])
def test_strip_tags_simple(input, expected):
    r = Richtypo()
    r.text = input
    r.strip_tags()

    expected = re.sub(r'~(\d+)~', '~([^~]+)~', expected)  # it is a regex now
    expected = re.compile(expected)
    assert expected.match(r.text)


@pytest.mark.parametrize("input", [
    '<h1>test</h1>',
    '<h1>тест</h1>',
    '<h1>test</',
    'тест',
    '<!-- h1 -->test</-->',
    '<H1>test</H1>',
    '<h1>test<h1>',
    '<h1><3>test</h1>',
    '<h1><10>test</h1>',
    '<p>test is a <a class="testclass">test link</a></p>',
])
def test_restore_tags(input):
    r = Richtypo()
    r.text = input

    r.strip_tags()
    r.restore_tags()
    assert r.text == input


@pytest.mark.parametrize("input, expected", [
    (
        '<script type="text/javascript">script_code()</script><h1>test</h1>',
        '~1~~2~test~3~'
    ),
    (
        """
        <script type="text/javascript">
            console.log("<script type="text/javascript">alert('test');</script>");
        </script>
        <h1>test<h2>
        """,

        """
        ~1~
        ~2~test~3~
        """
    ),
    (
        """
        <script type="text/javascript">
            scriptcode()
        </script>
        <h1>test</h1>
        """,

        """
        ~1~
        ~2~test~3~
        """
    ),
])
def test_strip_tags_in_scripts(input, expected):
    r = Richtypo()
    r.text = input
    r.strip_tags()

    expected = re.sub(r'~(\d+)~', '~([^~]+)~', expected)  # it is a regex now
    expected = re.compile(expected)
    assert expected.match(r.text)


@pytest.mark.parametrize("input", [
    '<script type="text/javascript">scriptcode()</script><h1>test</h1>',

    """
    <script type="text/javascript">
        console.log("<script type="text/javascript">alert('test');</script>");
    </script><h1>test<h2>
    """,

    """
    <script type="text/javascript">
            script_code()
    </script>
    <h1>test</h1>
    """,
    """
    <p>This is my fancy code:
        <div class="code">
            <code>Test();</code>
        </div>
    </p>
    """,
])
def test_restore_tags_in_scripts(input):
    r = Richtypo()
    r.text = input

    r.strip_tags()
    r.restore_tags()
    assert r.text == input


@pytest.mark.parametrize("input, expected", [
    ('test <h1> ~dff~', 'test <h1> ~dff~'),
    ('test <h1>~~dff~~</h1>', 'test <h1>~~dff~~</h1>'),
    ('test <h1> ~~dff~x~', 'test <h1> ~~dff~x~'),
])
def test_strip_tags_when_input_already_contains_richtypo_marks(input, expected):
    r = Richtypo()
    r.text = input
    r.strip_tags()
    r.restore_tags()
    assert r.text == expected



# -*- coding: utf-8
import pytest
import six

from richtypo import Richtypo
from richtypo.rules import SPECIAL_CHARACTERS_MAP


def _replace_unicode_back_to_special_characters(unicode_input):
    """
    Translate the output of richtypo to the language used in rule files,
    e.g. translate ` ` (non-breakable space) to `_`
    """
    for k, v in six.iteritems(SPECIAL_CHARACTERS_MAP):
        unicode_input = unicode_input.replace(v, k)

    return unicode_input


def test_rule_specs(rule_name, rule):
    """
    This test is a parametrizes spec checker for every rule defined in yaml
    """
    for spec in rule.specs:
        (input, expected) = spec.split('|')
        got = rule.apply(input)
        got = _replace_unicode_back_to_special_characters(got)

        assert got == expected, 'Rule: %s, Pattern: %s' % (rule_name, rule.pattern)


@pytest.mark.parametrize('input, expected', [
    ("Putin krab . Lenin grib.", "Putin krab. Lenin grib."),
    ("Putin krab . Lenin grib .", "Putin krab. Lenin grib."),
    ("Putin krab . Lenin grib . ", "Putin krab. Lenin grib."),
    ("Putin krab ! Lenin grib ! ", "Putin krab! Lenin grib!")
])
def test_punctuation(input, expected):
    """
    Test the complex rule for spaces after punctuation marks:
    the 'generic:punctuation_spaces' which should also be followed
    by 'generic:spaces_eol'
    """
    r = Richtypo()
    assert r.richtypo(input) == expected



# -*- coding: utf-8
from richtypo import Richtypo
from richtypo.rules import NBSP


def test_russian_simple():
    r = Richtypo(ruleset='ru-lite')
    text = u'Из-под топота копыт'
    assert r.richtypo(text) == u'Из-под%sтопота копыт' % NBSP


def test_unicode_with_ascii_only_characters():
    """
    If this test does not fail, and the above test does,
    then i've messed up with unicode
    """
    r = Richtypo(ruleset='en-lite')
    text = u'2 to 3'
    assert r.richtypo(text) == u'2%sto%s3' % (NBSP, NBSP)



from setuptools import find_packages, setup

setup(
    name='richtypo',
    version='0.2.3',
    description="The only web typograph that don't make a mess out of your text. Russian and English supported.",
    keywords=['typography', 'spaces', 'nbsp', 'russian', 'markdown', ],
    url="https://github.com/f213/richtypo.py",
    classifiers=[
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: Markup :: HTML',
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Natural Language :: Russian',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Framework :: Flask',
        'License :: OSI Approved :: MIT License',
    ],
    author="Fedor Borshev",
    author_email="f@f213.in",
    license="MIT",
    packages=find_packages(),
    setup_requires=['setuptools-git'],
    install_requires=[
        'six',
        'PyYAML',
        'backports.functools-lru-cache',
    ],
    include_package_data=True,
    zip_safe=False,
)



# -*- coding: utf-8
from __future__ import absolute_import

import re
from uuid import uuid4

import six

from richtypo import rules


class Richtypo(object):
    rulesets = {
        'generic': [
            'cleanup_before',
            'emdash-forced',
            'emdash-middle',
            'generic:nbsp',
            'generic:hanging_emdash',
            'generic:punctuation_spaces',
            'generic:spaces_eol',
            'generic:multiple_spaces',
        ],
        'ru-lite': [
            'cleanup_before',
            'emdash-forced',
            'emdash-middle',

            'generic:numbers',

            'ru:hanging_pretexts',
            'ru:hanging_unions',
            'ru:hanging_particles',
            'ru:hanging_pronouns',
            'ru:what_if',
            'ru:year',

            'generic:punctuation_spaces',
            'generic:spaces_eol',
            'generic:multiple_spaces',
            'generic:nbsp',
            'generic:hanging_emdash',
        ],
        'en-lite': [
            'cleanup_before',
            'emdash-forced',
            'emdash-middle',

            'en:hanging_pretexts',
            'en:hanging_articles',
            'en:hanging_pronouns',
            'en:hanging_to_be',
            'en:hanging_numbers',

            'generic:punctuation_spaces',
            'generic:spaces_eol',
            'generic:multiple_spaces',
            'generic:numbers',
            'generic:nbsp',
            'generic:hanging_emdash',
        ],
    }

    bypass_tags = [
        'script',
        'pre',
        'code'
    ]

    def __init__(self, ruleset='generic'):
        self.rules = []
        self.available_rules = {}

        self.init_tag_regexes()

        self.load_rules_for_ruleset(ruleset)
        self.build_rule_chain(ruleset)

    def richtypo(self, text):
        """
        Actualy do the stuff
        """
        self.text = text

        self.strip_tags()
        self.apply_rule_chain()
        self.restore_tags()

        return self.text

    @classmethod
    def _get_ruleset_rules(cls, ruleset):
        """
        Returns an iterable of rules, defined in the particular ruleset.

        Raises KeyError when ruleset is not defined.
        """
        return cls.rulesets[ruleset]

    @classmethod
    def _ruleset_files(cls, ruleset):
        """
        Returns an iterable with rule files, that are required to be loaded
        for the particular ruleset to work.
        """
        ruledefs = []
        for rule in cls._get_ruleset_rules(ruleset):
            if isinstance(rule, six.string_types):  # deal only with non-special rules, defined in YAML
                try:
                    (ruledef, rule_name) = rule.split(':')
                except ValueError:
                    ruledef = 'generic'

                if ruledef not in ruledefs:
                    ruledefs.append(ruledef)

                    yield ruledef

    @classmethod
    def init_tag_regexes(cls):
        """
        Build a list of regexps used by strip_tags()
        """
        cls.strip_tag_regexps = []

        cls.strip_tag_regexps.append(
            re.compile(r'(~+[^~]+~+)')  # strip all ~stuff~ occurences, because this mark is used by richtypo
        )

        for tag in cls.bypass_tags:
            cls.strip_tag_regexps.append(
                re.compile(r'(<%s[^>]*>.+</%s>)' % (tag, tag), flags=re.MULTILINE | re.DOTALL)  # add a regex for every tag in a list of tags to bypass
            )

        cls.strip_tag_regexps.append(
            re.compile(r'(<[^>]+>)')  # generic regex to strip all <tags>
        )

    def _get_rule(self, rule):
        """
        Finds the rule in the list of loaded rules. Raises KeyError if rule is not found.

        Rules are devided into two types — string name (like `'emdash'`) or a class name
        of rule defined in richtypo.rules (like `StripDashes`).

        String rules have prefixes of files, from which they were loaded, i.e. rule
        loaded from `rules/ru.yaml` and called `emdash` can be found by name 'ru:emdash'.
        There is only one exception — you can find rules loaded from `rules/generic.yaml`
        without the 'generic:' prefix, simply `'cleanup_before'`.
        """
        if isinstance(rule, six.string_types):
            try:
                return self.available_rules[rule]
            except KeyError:
                try:
                    return self.available_rules['generic:' + rule]
                except KeyError:
                    raise KeyError('Rule not found:', rule)

        else:
            if issubclass(rule, rules.Rule):
                return rule
            else:
                raise KeyError('Got unkown rule reference', rule)

    def strip_tags(self):
        """
        Replace all tags with ~<tag_num>~
        """
        self.saved_tags = {}

        def repl(m):
            tag = m.group(1)
            uuid = uuid4()
            self.saved_tags[str(uuid)] = tag

            return '~%s~' % uuid

        for regex in self.strip_tag_regexps:
            self.text = regex.sub(repl, self.text)

    def restore_tags(self):
        """
        Restore tags, stripped by strip_tags
        """
        restore_tags = re.compile(r'~(?P<saved_tag_id>[^~]+)~')  # find each tag uuid
        self.text = restore_tags.sub(lambda m: self.saved_tags.pop(m.group('saved_tag_id')), self.text)  # and replace it from self.saved_tags dict

    def apply_rule_chain(self):
        for rule in self.rules:
            self.text = rule.apply(self.text)

    def build_rule_chain(self, ruleset):
        """
        Build rule chain for a particular ruleset

        Will throw KeyError if ruleset is not found.
        """
        self.rules = [self._get_rule(r) for r in self._get_ruleset_rules(ruleset)]

    def load_rules_for_ruleset(self, ruleset):
        """
        Load rule files needed by the particular ruleset
        """
        for ruledef in self._ruleset_files(ruleset):
            self._load_rules_from_file(ruledef)

    def _load_rules_from_file(self, file):
        """
        Load rules from the rule file
        """
        for rule_name, rule in rules.load_from_file(file):
            rule_name = '%s:%s' % (file, rule_name)
            if self.available_rules.get(rule_name) is None:
                self.available_rules[rule_name] = rule



# -*- coding: utf-8
import re

import six

from richtypo import Richtypo
from richtypo.rules import ABRule, Rule

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


def test_rule_generic():
    r = Rule(
        pattern='b{2}',
        replacement='c'
    )
    assert r.apply('abb') == 'ac'


def test_rule_generic_no_match():
    r = Rule(
        pattern='b{2}',
        replacement='c'
    )
    assert r.apply('acc') == 'acc'


def test_rule_applying():
    r1 = Rule(
        pattern='b{2}',
        replacement='cc'
    )
    r2 = Rule(
        pattern='c{2}',
        replacement='dd'
    )

    r = Richtypo()
    r.rules = [r1, r2]
    r.text = 'abb'
    r.apply_rule_chain()

    assert r.text == 'add'


def test_rule_order():
    """
    Check if rules are applyied in order they are supplied
    """
    r1 = Rule(
        pattern='b{2}',
        replacement='cc'
    )
    r2 = Rule(
        pattern='c{2}',
        replacement='dd'
    )
    r = Richtypo()
    r.rules = [r2, r1]  # r2 works only after r1
    r.text = 'abb'
    r.apply_rule_chain()

    assert r.text == 'acc'  # so the text should not be changed


def test_get_rule_from_available():
    r = Richtypo()
    r.available_rules = {
        'b': Rule(pattern='b', replacement='d')
    }
    rule = r._get_rule('b')

    assert rule.pattern == 'b'


def test_get_rule_from_predefined_rules():
    r = Richtypo()
    rule = r._get_rule(ABRule)

    assert rule == ABRule


def test_rule_flags():
    rule = Rule(pattern='A', replacement='b')
    rule._compile()

    if six.PY2:
        reference_flags = re.UNICODE
    else:
        reference_flags = 0

    assert rule._re == re.compile('A', flags=reference_flags)

    rule.flags = ['I']
    rule._compile()
    assert rule._re == re.compile('A', flags=reference_flags | re.I)


def test_all_occurancies_are_replaced():
    rule = Rule(pattern='(a|b)', replacement=r'\1c')
    assert rule.apply('aab') == 'acacbc'


@patch('richtypo.Richtypo._get_ruleset_rules')
def test_build_rule_chain(rules):
    r = Richtypo()
    r.available_rules = {
        'b': Rule(pattern='b', replacement='d')
    }

    with patch('richtypo.Richtypo._get_ruleset_rules') as rules:
        rules.return_value = ['b', ABRule]

        r.build_rule_chain('generic')
        assert len(r.rules) == 2


@patch('richtypo.Richtypo._get_ruleset_rules')
def test_ruleset_input_param(get_ruleset):
    get_ruleset.return_value = [
        ABRule,
    ]
    r = Richtypo(ruleset='generic')  # this ruleset realy should exist
    assert len(r.rules) == 1
    assert ABRule in r.rules


def test_rule_loading_by_default():
    r = Richtypo()
    assert len(r.available_rules.keys()) >= 1
